import functools
import pika
import time

from varys.utils import varys_message
from varys.process import Process


class Consumer(Process):
    def __init__(
        self,
        message_queue,
        exchange,
        configuration,
        log_file,
        log_level,
        queue_suffix,
        exchange_type,
        prefetch_count,
        routing_key="arbitrary_string",
        reconnect_wait=10,
    ):
        super().__init__(
            message_queue,
            exchange,
            configuration,
            log_file,
            log_level,
            queue_suffix,
            exchange_type,
            routing_key=routing_key,
            reconnect_wait=reconnect_wait,
        )

        self._closing = False
        self._prefetch_count = prefetch_count

    def _on_message(self, _unused_channel, basic_deliver, properties, body):
        message = varys_message(basic_deliver, properties, body)
        self._log.info(
            f"Received Message: #{message.basic_deliver.delivery_tag} from {message.properties.app_id}, {message.body}"
        )
        self._message_queue.put(message)

    def _acknowledge_message(self, delivery_tag):
        self._log.info(f"Acknowledging message: {delivery_tag}")
        self._connection.add_callback_threadsafe(
            functools.partial(
                self._channel.basic_ack,
                delivery_tag=delivery_tag,
            )
        )

    def _nack_message(self, delivery_tag, requeue):
        self._log.info(f"Nacking message: {delivery_tag}")
        self._connection.add_callback_threadsafe(
            functools.partial(
                self._channel.basic_nack,
                delivery_tag=delivery_tag,
                multiple=False,
                requeue=requeue,
            )
        )

    def run(self):
        while not self._stopping:
            try:
                # clear the queue, otherwise acking can cause problems on new channel
                while not self._message_queue.empty():
                    self._message_queue.get()

                self._connection = pika.BlockingConnection(self._parameters)
                self._channel = self._connection.channel()
                self._channel.exchange_declare(
                    exchange=self._exchange,
                    exchange_type=self._exchange_type,
                    durable=True,
                )
                self._channel.queue_declare(queue=self._queue, durable=True)
                self._channel.queue_bind(
                    queue=self._queue,
                    exchange=self._exchange,
                    routing_key=self._routing_key,
                )
                self._channel.basic_qos(prefetch_count=self._prefetch_count)
                self._channel.basic_consume(
                    self._queue, self._on_message, auto_ack=False
                )
                self._channel.start_consuming()
            except Exception as e:
                self._log.exception("Consumer caught exception:")

            if self._stopping or self._reconnect_wait < 0:
                break
            else:
                time.sleep(self._reconnect_wait)
                continue

    def stop(self):
        self._log.info("Stopping consumer as instructed...")
        self._stopping = True

        self._connection.add_callback_threadsafe(self._channel.stop_consuming)

        self._connection.add_callback_threadsafe(self._channel.close)

        self._connection.add_callback_threadsafe(self._connection.close)

        self._log.debug("Stopping consumer logger...")
        self._stop_logger()

        self._log.info("Stopped consumer as instructed.")



import unittest
import time
import tempfile
import os
import json
import logging
from varys import Varys
import pika

DIR = os.path.dirname(__file__)
LOG_FILENAME = os.path.join(DIR, "test.log")
TMP_HANDLE, TMP_FILENAME = tempfile.mkstemp()
TEXT = "Hello, world!"


class TestVarys(unittest.TestCase):

    def tearDown(self):
        # this seems to prevent some hanging
        # or errors related to closing connections that haven't opened yet
        # I presume because some operations are so fast
        # that we try to close the connections before they've opened
        # 0.01s seems to be sufficient; 0.1s is just a bit conservative
        time.sleep(0.1)

        self.v.close()
        os.remove(TMP_FILENAME)
        time.sleep(0.1)

        credentials = pika.PlainCredentials("guest", "guest")

        connection = pika.BlockingConnection(
            pika.ConnectionParameters("localhost", credentials=credentials)
        )
        channel = connection.channel()

        channel.queue_delete(queue="test_varys.q")

        connection.close()

        # check that all file handles were dropped
        logger = logging.getLogger("test_varys")
        self.assertEqual(len(logger.handlers), 0)

    def send_and_receive(self):
        self.v.send(TEXT, "test_varys", queue_suffix="q")
        message = self.v.receive("test_varys", queue_suffix="q")
        self.assertEqual(TEXT, json.loads(message.body))

        logger = logging.getLogger("test_varys")
        self.assertEqual(len(logger.handlers), 1)

    def manual_ack(self):

        self.v.auto_ack = False

        self.v.send(TEXT, "test_varys", queue_suffix="q")

        message = self.v.receive("test_varys", queue_suffix="q")

        self.v.acknowledge_message(message)

    def nack(self):
        self.v.auto_ack = False

        self.v.send(TEXT, "test_varys", queue_suffix="q")

        message = self.v.receive("test_varys", queue_suffix="q")

        self.v.nack_message(message)

        # check that the message has been requeued
        message_2 = self.v.receive("test_varys", queue_suffix="q")

        self.v.acknowledge_message(message_2)

        self.assertEqual(message.body, message_2.body)

    def send_and_receive_batch(self):
        self.v.send(TEXT, "test_varys", queue_suffix="q")
        self.v.send(TEXT, "test_varys", queue_suffix="q")

        messages = self.v.receive_batch("test_varys", queue_suffix="q", timeout=1)
        parsed_messages = [json.loads(m.body) for m in messages]
        self.assertListEqual([TEXT, TEXT], parsed_messages)

    def receive_no_message(self):
        self.assertIsNone(
            self.v.receive("test_varys", queue_suffix="q", timeout=0)
        )

    def send_no_suffix(self):
        self.assertRaises(Exception, self.v.send, TEXT, "test_varys")

    def receive_no_suffix(self):
        self.assertRaises(Exception, self.v.receive, "test_varys")

    def receive_batch_no_suffix(self):
        self.assertRaises(Exception, self.v.receive_batch, "test_varys")


class TestVarysTLS(TestVarys):

    def setUp(self):
        config = {
            "version": "0.1",
            "profiles": {
                "test": {
                    "username": "guest",
                    "password": "guest",
                    "amqp_url": "localhost",
                    "port": 5671,
                    "use_tls": True,
                    "ca_certificate": ".rabbitmq/ca_certificate.pem",
                }
            },
        }

        with open(TMP_FILENAME, "w") as f:
            json.dump(config, f, ensure_ascii=False)

        self.v = Varys("test", LOG_FILENAME, config_path=TMP_FILENAME)

    def test_send_and_receive(self):
        self.send_and_receive()

    def test_manual_ack(self):
        self.manual_ack()

    def test_nack(self):
        self.nack()

    def test_send_and_receive_batch(self):
        self.send_and_receive_batch()

    def test_receive_no_message(self):
        self.receive_no_message()

    def test_send_no_suffix(self):
        self.send_no_suffix()

    def test_receive_no_suffix(self):
        self.receive_no_suffix()

    def test_receive_batch_no_suffix(self):
        self.receive_batch_no_suffix()


class TestVarysNoTLS(TestVarys):

    def setUp(self):
        config = {
            "version": "0.1",
            "profiles": {
                "test": {
                    "username": "guest",
                    "password": "guest",
                    "amqp_url": "127.0.0.1",
                    "port": 5672,
                    "use_tls": False,
                    "ca_certificate": "this-value-shouldn't-matter",
                }
            },
        }

        with open(TMP_FILENAME, "w") as f:
            json.dump(config, f, ensure_ascii=False)

        self.v = Varys("test", LOG_FILENAME, config_path=TMP_FILENAME)

    def test_send_and_receive(self):
        self.send_and_receive()

    def test_manual_ack(self):
        self.manual_ack()

    def test_nack(self):
        self.nack()

    def test_send_and_receive_batch(self):
        self.send_and_receive_batch()

    def test_receive_no_message(self):
        self.receive_no_message()

    def test_send_no_suffix(self):
        self.send_no_suffix()

    def test_receive_no_suffix(self):
        self.receive_no_suffix()

    def test_receive_batch_no_suffix(self):
        self.receive_batch_no_suffix()


class TestVarysConfig(unittest.TestCase):
    def tearDown(self):
        os.remove(TMP_FILENAME)

    def test_config_not_json(self):
        with open(TMP_FILENAME, "w") as f:
            f.write("asdf9υ021ζ3;-ö×=()[]{}∇Δοo")

        # use a context manager so we can check SystemExit code
        with self.assertRaises(SystemExit) as cm:
            v = Varys("test", LOG_FILENAME, config_path=TMP_FILENAME)

        self.assertEqual(cm.exception.code, 11)

    def test_config_profile_missing(self):
        config = {
            "version": "0.2",  # bad version prints warning but doesn't raise error
            "profiles": {"asdfadsf": {}},
        }

        with open(TMP_FILENAME, "w") as f:
            json.dump(config, f, ensure_ascii=False)

        with self.assertRaises(SystemExit) as cm:
            v = Varys("test", LOG_FILENAME, config_path=TMP_FILENAME)

        self.assertEqual(cm.exception.code, 2)

    def test_config_profile_incomplete(self):
        config = {
            "version": "0.1",
            "profiles": {
                "test": {
                    "username": "username",
                    "extra": "unnecessary",
                }
            },
        }

        with open(TMP_FILENAME, "w") as f:
            json.dump(config, f, ensure_ascii=False)

        with self.assertRaises(SystemExit) as cm:
            v = Varys("test", LOG_FILENAME, config_path=TMP_FILENAME)

        self.assertEqual(cm.exception.code, 11)

if __name__ == '__main__':
    unittest.main()



import queue
import os
import time

from varys.consumer import Consumer
from varys.producer import Producer
from varys.utils import configurator


class Varys:
    """
    A high-level wrapper for the producer and consumer classes used by varys, abstracting away the tedious details.

    ...

    Attributes
    ----------
    profile : str
        profile name inside the configuration file to use when connecting to RabbitMQ
    configuration_path : str
        path to varys confiruration JSON file, provided with either the config_path argument or the VARYS_CFG environment variable
    _logfile : str
        the path to the logfile to use for logging, provided with the logfile argument
    _log_level : str
        the log level to use for logging, provided with the log_level argument, defaults to DEBUG (the most verbose logging level)
    _credentials : class
        an instance of the configurator class, used to store the RabbitMQ connection credentials
    _in_channels : dict
        a dictionary of consumer classes and queues that have been connected to for receiving messages
    _out_channels : dict
        a dictionary of producer classes and queues that have been connected to for sending messages

    Methods
    -------
    send(message, exchange, queue_suffix=False, exchange_type="fanout")
        Either send a message to an existing exchange, or create a new exchange connection and send the message to it. queue_suffix must be provided when sending a message to a queue for the first time to instantiate a new connection.
    receive(exchange, queue_suffix=False, block=True, timeout=None, exchange_type="fanout", prefetch_count=5)
        Either receive a message from an existing exchange, or create a new exchange connection and receive a message from it. queue_suffix must be provided when receiving a message from a queue for the first time to instantiate a new connection. block determines whether the receive method should block until a message is received or not.
    receive_batch(exchange, queue_suffix=False)
        Either receive a batch of messages from an existing exchange, or create a new exchange connection and receive a batch of messages from it. queue_suffix must be provided when receiving a message from a queue for the first time to instantiate a new connection.
    get_channels()
        Return a dict of all the channels that have been connected to with the keys "consumer_channels" and "producer_channels"
    close()
        Close all open channels
    """

    def __init__(
        self,
        profile,
        logfile,
        log_level="DEBUG",
        config_path=None,
        routing_key="arbitrary_string",
        auto_acknowledge=True,
    ):
        self.profile = profile

        if config_path is None:
            config_path = os.getenv("VARYS_CFG")

        self.configuration_path = config_path

        self.routing_key = routing_key
        self.auto_ack = auto_acknowledge

        self._logfile = logfile
        self._log_level = log_level

        self._credentials = configurator(self.profile, self.configuration_path)

        self._in_channels = {}
        self._out_channels = {}

    def send(
        self,
        message,
        exchange,
        queue_suffix=False,
        exchange_type="fanout",
        max_attempts=1,
    ):
        """
        Either send a message to an existing exchange, or create a new exchange connection and send the message to it.
        """

        if not self._out_channels.get(exchange):
            if not queue_suffix:
                raise Exception(
                    "Must provide a queue suffix when sending a message to a queue for the first time"
                )

            self._out_channels[exchange] = Producer(
                message_queue=queue.Queue(),
                routing_key=self.routing_key,
                exchange=exchange,
                configuration=self._credentials,
                log_file=self._logfile,
                log_level=self._log_level,
                queue_suffix=queue_suffix,
                exchange_type=exchange_type,
            )
            self._out_channels[exchange].start()
            time.sleep(0.1)

        self._out_channels[exchange].publish_message(message, max_attempts=max_attempts)

    def receive(
        self,
        exchange,
        queue_suffix=False,
        timeout=None,
        exchange_type="fanout",
        prefetch_count=5,
    ):
        """
        Either receive a message from an existing exchange, or create a new exchange connection and receive a message from it.
        """

        if not self._in_channels.get(exchange):
            if not queue_suffix:
                raise Exception(
                    "Must provide a queue suffix when receiving a message from an exchange for the first time"
                )

            self._in_channels[exchange] = Consumer(
                message_queue=queue.Queue(),
                routing_key=self.routing_key,
                exchange=exchange,
                configuration=self._credentials,
                log_file=self._logfile,
                log_level=self._log_level,
                queue_suffix=queue_suffix,
                exchange_type=exchange_type,
                prefetch_count=prefetch_count,
            )
            self._in_channels[exchange].start()
            time.sleep(0.1)

        try:
            message = self._in_channels[exchange]._message_queue.get(
                block=True, timeout=timeout
            )
            if self.auto_ack:
                # Only ack a message when it is pulled out of the thread-safe queue and auto_ack is set
                self._in_channels[exchange]._acknowledge_message(
                    message.basic_deliver.delivery_tag
                )
            return message
        except queue.Empty:
            return None

    def receive_batch(
        self, exchange, queue_suffix=False, timeout=0, exchange_type="fanout"
    ):
        """
        Either receive all messages available from an existing exchange, or create a new exchange connection and receive all messages available from it.
        """
        if timeout is None:
            raise ValueError(
                "Timeout cannot be `None` or `receive_batch` would block forever."
            )

        messages = []
        while True:
            messages.append(
                self.receive(
                    exchange,
                    queue_suffix=queue_suffix,
                    timeout=timeout,
                    exchange_type=exchange_type,
                )
            )
            if messages[-1] is None:
                messages.pop()
                break

        return messages

    def acknowledge_message(self, message):
        """
        Acknowledge a message manually. Not necessary by default where auto_acknowledge is set to True.
        """

        self._in_channels[message.basic_deliver.exchange]._acknowledge_message(
            message.basic_deliver.delivery_tag
        )

    def nack_message(self, message, requeue=True):
        """
        Nack a message manually. Can be used to requeue a message if auto_acknowledge is set to False.

        """

        if self.auto_ack:
            raise Exception(
                "Cannot nack a message when auto_acknowledge is set to True"
            )

        self._in_channels[message.basic_deliver.exchange]._nack_message(
            delivery_tag=message.basic_deliver.delivery_tag, requeue=requeue
        )

    def get_channels(self):
        """Return all open channels."""

        return {
            "consumer_channels": self._in_channels.keys(),
            "producer_channels": self._out_channels.keys(),
        }

    def close(self):
        """Close all open channels."""

        for channel in self._in_channels.values():
            channel.stop()
            channel.join()

        for channel in self._out_channels.values():
            channel.stop()
            channel.join()
